"""
Utility functions for S3.
"""

import pandas as pd
import requests
import s3fs

s3 = s3fs.S3FileSystem(anon=True)  # All on public bucket

S3_PUBLIC_URL_BASE = "https://aind-scratch-data.s3.us-west-2.amazonaws.com/aind-patchseq-data"
S3_PATH_BASE = "aind-scratch-data/aind-patchseq-data"


def check_s3_public_url_exists(s3_url: str) -> bool:
    """Check if a given s3 url exists."""
    response = requests.get(s3_url)
    return response.status_code == 200


def get_public_url_sweep(ephys_roi_id: str, sweep_number: int) -> str:
    """Get the public URL for a sweep."""

    s3_sweep = (
        f"{S3_PUBLIC_URL_BASE}/efel/plots/{ephys_roi_id}/{ephys_roi_id}_sweep_{sweep_number}.png"
    )
    s3_spikes = (
        f"{S3_PUBLIC_URL_BASE}/efel/plots/{ephys_roi_id}/"
        f"{ephys_roi_id}_sweep_{sweep_number}_spikes.png"
    )

    # Check if the file exists on s3 public
    urls = {}
    if check_s3_public_url_exists(s3_sweep):
        urls["sweep"] = s3_sweep
    if check_s3_public_url_exists(s3_spikes):
        urls["spikes"] = s3_spikes
    return urls


def get_public_efel_cell_level_stats():
    """Get the cell level stats that are generated by eFEL (and merged with the metadata)."""
    csv_url = f"{S3_PUBLIC_URL_BASE}/efel/cell_stats/cell_level_stats.csv"
    if check_s3_public_url_exists(csv_url):
        return pd.read_csv(csv_url)
    else:
        raise FileNotFoundError(f"Cell level stats CSV file not found at {csv_url}")


def get_public_url_cell_summary(ephys_roi_id: str, if_check_exists: bool = True) -> str:
    """Get the public URL for a cell summary plot."""
    s3_url = f"{S3_PUBLIC_URL_BASE}/efel/cell_stats/{ephys_roi_id}_cell_summary.png"
    if if_check_exists:
        if check_s3_public_url_exists(s3_url):
            return s3_url
        else:
            return None
    else:
        return s3_url


if __name__ == "__main__":
    # print(get_public_url_sweep("1212546732", 46))
    print(get_public_efel_cell_level_stats())
